/*
 * Copyright 1993-2002 Christopher Seiwald and Perforce Software, Inc.
 *
 * This file is part of Jam - see jam.c for Copyright information.
 */

/* Keep JAMVERSYM in sync with VERSION. */
/* It can be accessed as $(JAMVERSION) in the Jamfile. */

#define VERSION_MAJOR 2015
#define VERSION_MINOR 7
#define VERSION_PATCH 0
#define VERSION_MAJOR_SYM "2015"
#define VERSION_MINOR_SYM "07"
#define VERSION_PATCH_SYM "00"
#define VERSION "2015.07"
#define JAMVERSYM "JAMVERSION=2015.07"
